#include <string.h>
#include "spins.h"
#include "symmetry.h"
#include "config.h"
/*
 *  - read lattice data from file (line by line) -> MODELFILE=daten.def 
 *   03/00 numpara dynamisch, ww.w added (twisted boundary conditions)
 *   04/02 new (compatible) format allows optional parameters
 *         of the format <char><idx>, ex: J1 t2 u4
 *   07/01 four-site-terms 1st line nq=8 3th data-block idx,s1..s4,K5
 */
/* ================== private-part =========================== */
#define MODELFILE "daten.def"

double pos[NN][3];             /* site-position = (x,y,z)      */
double edg[3][3];              /* edge vectors not_given=0 */
int  iw[Nw][2];                /* array bonds                  */
int  iq[Nq][4];                /* array four-site-terms (quad?) */

struct tww ww;                 /* Parameter H                  */

int  nn,nw,nq;                 /* number of spins and bonds NN,Nw,Nq */
int  nu,nd;                    /* number up-/down-spins        */
char numpara=0;                /* array for values     */
double para[MaxPara];          /* values for wt,wj,etc. */


int idx[NN];

/*  i=site 0..NN, d=direction 0..2, s=steps ..,-1,0,1,.. rc=i+s*vec_d */
/* better site(i,dx,dy[,dz])  dy<0 dont work !!!!  */
int site_move(int i,int dx,int dy,...){
  if(i<0 || i>=NN)return -1;
  if(dx==0 && dy==0) return 0;
  return(i);
}

/*
 *  # +++model+++
 *  # NN Nw (Nv is obsolete)
 *    2  1  [nq=<number quad terms>]
 *  # s  X  Y  Z  e  U
 *    0  0  0  0  0  0
 *    1  1  0  0  0  0
 *  # n s1 s2  t  J  V         old style
 *    0  0  1  2  1  0       # old style
 *  # n s1 s2   [R= %d %d %d] [t%d] [J%d] [V%d]     new style
 *    0  0  1    R=  1  0  0    t2    J1    V0      new style
 *  # n s1 s2 s3 s4 [R= ??? ] [Q%d] (optional, nq!=0)
 *    0  0  1  2  3 Q5
 *  # ---model---
 */

#include <ctype.h>

/* parser added for better inputfile error-detection  Mar2000
 * count the number of numeric strings, stop for comments and a-zA-Z
 */
int sdata(char *s){ /* check the string for usefull numeric data */
    int i,j,mode=0;
    for(i=j=0;i<1024 && s[i]!=0;i++){
      if( s[i]=='#' || s[i]=='%' || s[i]=='!'
       || s[i]==';' || isalpha(s[i])) break; /* comment or new-style */
      if( s[i]==' ' || s[i]=='\t' || s[i]=='\r') { mode=0; } /* space */
      else { if(!mode){ mode=1;j++; } }
    }
    return j;	/* return number of values found */
}

/* parse a line (s) for a specifique parameter string (cc)
 * return its index
 * parser for new style: example idata(" 0 1 2 j2 u1 t5","j")=2
 * can also be used to find "nq=" and return the argument
 */
int idata(char *s, char *cc){
    int i,j,mode=0;
    for(i=0;i<1024 && s[i]!=0;i++){
      if(s[i]=='#' || s[i]=='%' || s[i]=='!' || s[i]==';') break; /* comment */
      if(s[i]==' ' || s[i]=='\t' || s[i]=='\r') { mode=0; } /* space */
      else {
        if(!mode){ mode=1; /* nospace */
          for (j=0; cc[j] && s[i+j] && tolower(s[i+j])==tolower(cc[j]); j++);
          if (!cc[j]) {
            /* printf("\n dbg: c=%c i=%d s=%s",cc,atoi(s+i+1),s); */
            /* found the string, convert the number which may follow */
            return atoi(s+i+j);
          }
        } 
      }
    }
    return 0;	/* not found */
}

int initvars(){	/* function definition */
    int i,j,l,p_max=0,nv=0; FILE *f1; char s1[1024];
    for (i=0;i<3;i++) for (j=0;j<3;j++) edg[i][j]=0; /* default */
    nn=NN;nw=Nw;nu=NN/2;nd=NN/2+NN%2; /* number of up- and down-spins */
    nq=0; /* default is no quad terms */
    f1=fopen(MODELFILE,"rt");
    i=j=0; l=0; s1[0]=0;              /* l: line number, s1: line */
    if(!f1)F1("could not open file %s",MODELFILE);
    if(vvv)printf(" read %s\n",MODELFILE);
    do {
      if (!fgets(s1,1023,f1)) F0("read"); l++;
      if (s1[0]=='#' && l<8 && vvv) printf("%s",s1); /* output header */
      if (strncmp(s1,"#===edge===",11)==0) { /* edgevectors given? */
        i=sdata(s1+11); /* test for dimension = 1, 2 or 3 */
        if(i==1) sscanf(s1+11,"%lf",&edg[0][0]);
        if(i==4) sscanf(s1+11,"%lf%lf%lf%lf",
                   &edg[0][0],&edg[0][1],
                   &edg[1][0],&edg[1][1]);
        if(i==9) sscanf(s1+11,"%lf%lf%lf%lf%lf%lf%lf%lf%lf",
                   &edg[0][0],&edg[0][1],&edg[0][2],
                   &edg[1][0],&edg[1][1],&edg[1][2],
                   &edg[2][0],&edg[2][1],&edg[2][2]);
      }
    } while(!sdata(s1));
    if(sdata(s1)<2)F1("read line %d: numbers NN and Nw were expected",l);
    sscanf(s1,"%d%d",&nn,&nw); nq=idata(s1,"nq=");
    if(NN!=nn || Nw<nw){ 
      printf("\nWARNING: NN Nw not unique!"
             "\n daten.def: nn= %3d nw= %3d"
             "\n config.h:  NN= %3d Nw= %3d\n",nn,nw,NN,Nw);
    }
    /* NN, Nw and Nq are used to define array boundaries, need a check */
    if (NN<nn || Nw<nw || Nq<nq)
      F0("NN Nw or Nq to small for daten.def, change config.h");
    for(i=0;i<nn;i++){
      do { if(!fgets(s1,1023,f1))F0("read");l++; } while(!sdata(s1));
      ww.we[i]=ww.wu[i]=ww.wf[i]=ww.wd[i]=0; /* default: not set */
      if (sdata(s1)<5) 
        F1("read line %d: site Rx Ry Rz p_e p_U [Jx,Ux,Fx,Dx] expected",l);
      sscanf(s1,"%d%lf%lf%lf",&idx[i],
        &pos[i][0],&pos[i][1],&pos[i][2]); /* ToDo: subject of change R=(x,y,z) */
#if CM(0,1,1)
      sscanf(s1,"%*d%*f%*f%*f%d%d",
      	  &(ww.we[i]),&(ww.wu[i])); /* old format */
      j=idata(s1,"e"); if (j) ww.we[i]=j; /* also scanf(" e%d",..); */
      j=idata(s1,"u"); if (j) ww.wu[i]=j;
      if(ww.we[i]>p_max) p_max=ww.we[i];
      if(ww.wu[i]>p_max) p_max=ww.wu[i];
#endif
#ifdef HFIELD
      j=idata(s1,"f"); if (j) ww.wf[i]=j;
      if(ww.wf[i]>p_max) p_max=ww.wf[i];
#endif
#ifdef SSANISO
      j=idata(s1,"d"); if (j) ww.wd[i]=j;
      if(ww.wd[i]>p_max) p_max=ww.wd[i];
#endif
      /* ToDo: introduce spin-length as parameter? */
    }
    for(i=nn;i<NN;i++){  /* fill unused elements with zeros */
      idx[i]=0; pos[i][0]=pos[i][1]=pos[i][2]=0;
      ww.we[i]=ww.wu[i]=ww.wf[i]=0;
    }
    printf("start reading %d two-site-terms at line %d\n", nw, l+1);
    for(i=0;i<nw;i++){
      ww.w[i]=ww.wj[i]=ww.wt[i]=ww.wv[i]=ww.wh[i]=0;
      do { if(!fgets(s1,1023,f1))F0("read");l++; } while(!sdata(s1));
      if(sdata(s1)<5)F1("read line %d: idx site1 site2 p_t p_j [p_V] [p_w] [...]",l);
      /* read old style, obsolete soon */
      sscanf(s1,"%*u%d%d%d%d",&iw[i][0],&iw[i][1],&ww.wt[i],&ww.wj[i]);
      /* ToDo: sscanf(s1,"%*s R= %lf %lf %lf",&dist[0],&dist[1],&dist[2]); */
#ifdef CONFIG_TBC
      sscanf(s1,"%*u%*d%*d%*d%*d%*d%d",&ww.w[i]); /* angle in units of pi */
      j=idata(s1,"w"); if (j) ww.w[i]=j;   /* field or boundary angle */
      if(ww.w[i]>p_max) p_max=ww.w[i];
#endif
      j=idata(s1,"j"); if (j) ww.wj[i]=j;  /* exchange */
      if(ww.wj[i]>p_max) p_max=ww.wj[i];
#if CM(0,1,1)
      sscanf(s1,"%*u%*d%*d%*d%*d%d%*d",&ww.wv[i]); /* old style */
      j=idata(s1,"t"); if (j) ww.wt[i]=j;  /* hopping */
      /* nv is set to trigger a warning message */
      j=idata(s1,"v"); if (j) ww.wv[i]=j;  /* repulsion */       if (j) nv++;
      j=idata(s1,"h"); if (j) ww.wh[i]=j;  /* hole-repulsion */
      if(ww.wt[i]>p_max) p_max=ww.wt[i];
      if(ww.wv[i]>p_max) p_max=ww.wv[i];
      if(ww.wh[i]>p_max) p_max=ww.wh[i];
#endif
      /* printf("read line %d: iw[%d] = %s", l, j, s1); */
    }
    /* new part for four-site terms, Jan07 */
    printf("start reading %d four-site-terms at line %d\n", nq, l+1);
    for(i=0;i<nq;i++){
      ww.wq[i]=0;
      do { if(!fgets(s1,1023,f1))F1("read nq line %d",l); l++;
         } while(!sdata(s1));
      if(sdata(s1)<5)F1("read line %d: idx site1 2 3 4 [p_Q]",l);
      sscanf(s1,"%*u%d%d%d%d",&iq[i][0],&iq[i][1],&iq[i][2],&iq[i][3]);
      j=idata(s1,"q"); if (j) ww.wq[i]=j;  /* four-site term */
      if(ww.wq[i]>p_max) p_max=ww.wq[i];
      /* printf("read line %d: iq[%d] = %s", l, j, s1); */
    }
    if (nq>0) printf(
      "\n * * * Warning - no symmetry check for four-site-exchanges * * *\n\n");

    fclose(f1);
    if (nv) E0("V!=0 not checked yet, please check and inform the author");
    /* if(vvv)printf("\n"); */
    if (p_max+1>=MaxPara) E0("MaxPara exceeded (look at spins.h?)");
    numpara=p_max+1;		/* number of different values */
    para[0]= 0.0;	/* must be always 0 */
    para[1]= 1.0;	/* J1 (example) */
    para[2]= 0.4;	/* J2 */
    para[3]= 0.0;	/* J3 */
    para[4]=-1.0;	/* t4 */
    para[5]= 0.1;	/* t5 */
    para[6]=-0.2;	/* t6 */
    para[7]= 0.0;	/* e7 */
    para[8]= 8.0;	/* U8 */
    return 0;
}


/* obsolete, should be replaced by a better readable one */
int outmodel(){		/* never used ??? example! */
  FILE *f1; int i;
  /* output model in most simple form */
  f1=fopen("daten.def","wt");		 /* create new DemoFile */
  if(!f1)F0("create daten.def");	if(vvv)printf(" write daten.def");
  fprintf(f1,"#+++model+++\n");
  fprintf(f1,"# NN=Number_of_sites Nw=Number_of_bonds (Nv is obsolete)\n");
  fprintf(f1,"%2u %3u \n",nn,nw);
  fprintf(f1,"# s  X  Y  Z p_e p_U (s=site  R=(X Y Z)  p=parameter-idx)\n");
  for(i=0;i<NN;i++)
    fprintf(f1," %2u %2g %2g %2g %2d %2d f%2d\n",
    i,pos[i][0],pos[i][1],pos[i][2],ww.we[i],ww.wu[i],ww.wf[i]);
  fprintf(f1,"# Bond s1 s2 p_t p_J p_V p_w\n");
  for(i=0;i<nw;i++)
    fprintf(f1," %2u %2d %2d %2d %2d %2d %2d\n",
    i,iw[i][0],iw[i][1],ww.wt[i],ww.wj[i],ww.wv[i],ww.w[i]);
  fprintf(f1,"#---model---\n");
  fclose(f1);
  return(0);
}

