#include "defs.h"
#include "block.h"
#include "array_m.h"
#include "vector.h"
#include "hilbert.h"
#include "spins.h"
#ifdef CONFIG_PTHREAD
#include <pthread.h>
#endif

struct vblock {     /* arguments for working threads */
  int blk; mcplx *v0; /* block, vector (used by clrvec,mltply,addav) */
  double r;         /* factor (used by mltply,addav) */
  mcplx *v1;        /* vector (used by addav) */
};

/* ToDo: parallelize add_av, mltly_av and norm */

/* void-function for pthreads */
void add_av_block( void *ptr ){
  int   blk=((struct vblock *)ptr)->blk;
  mcplx *v0=((struct vblock *)ptr)->v0;
  mcplx *v1=((struct vblock *)ptr)->v1;
  double  r=((struct vblock *)ptr)->r, n2t=0;
  ULong i;
  for(i=0l;i<b_len[blk];i++) {
    v1[i] += M_d2f(M_f2d(v0[i])*r);
    n2t   += Norm(M_f2d(v1[i]));
  }
  ((struct vblock *)ptr)->r = n2t; /* scalar product (v1,v1) */
  return;
}
double add_av(tvec *v1, double r, tvec *v0){  /* v1+=r*v0, return norm(v1) */
  struct vblock buf[B_NUM];
  int blk; double n2=0;
  for(blk=0;blk<B_NUM;blk++) {
    buf[blk].blk=blk;
    buf[blk].r  =r;
    buf[blk].v0 =v0->v[blk]; /* vectorblock */
    buf[blk].v1 =v1->v[blk]; /* vectorblock */
    parallel_start( blk, add_av_block, &buf[blk]);
  }
  for (blk=0;blk<B_NUM;blk++) {
    parallel_wait(blk); /* synchronize */
    n2 += buf[blk].r;
  }
  return(n2);
}

/* void-function for pthreads */
void mltply_av_block( void *ptr ){
  int   blk=((struct vblock *)ptr)->blk;
  mcplx *v0=((struct vblock *)ptr)->v0;
  double  r=((struct vblock *)ptr)->r;
  ULong i;
  for(i=0l;i<b_len[blk];i++) 
    v0[i] = M_d2f(M_f2d(v0[i])*r);
  return;
}
void mltply_av(double r, tvec *v0){  /* v0*=r */
  struct vblock buf[B_NUM];
  int blk;
  for(blk=0;blk<B_NUM;blk++) {
    buf[blk].blk=blk;
    buf[blk].r  =r;
    buf[blk].v0 =v0->v[blk]; /* vectorblock */
    parallel_start( blk, mltply_av_block, &buf[blk]);
  }
  for (blk=0;blk<B_NUM;blk++) parallel_wait(blk); /* synchronize */
}

/* void-function for pthreads */
void clrvec_block( void *ptr ){
  int    j=((struct vblock *)ptr)->blk;
  mcplx *v=((struct vblock *)ptr)->v0;
  ULong i;
  for (i=0l;i<b_len[j];i++) v[i]=M_d2f(0);
  return;
}
void clrvec(tvec *v){ int j;
  /* its important to inititialize memory locally in the threads */
  struct vblock buf[B_NUM];
  for (j=0;j<B_NUM;j++) {
    buf[j].blk=j;
    free(v->v[j]); v->v[j]=(mcplx *)malloc(b_len[j]*Csize); /* trick for ccNUMA */
    buf[j].v0 =v->v[j]; /* vectorblock */
    parallel_start( j, clrvec_block, &buf[j]);
  }
  for (j=0;j<B_NUM;j++) parallel_wait(j); /* synchronize */
  return;
}

double normvec(tvec *v){
    double r=0.0; ULong j; int i;
    for(i=0;i<B_NUM;i++)for(j=0l;j<b_len[i];j++)r+=Norm(M_f2d(v->v[i][j]));
    if(r<=1e-20){if(vvv&4)printf(" norm_vec<1e-20");} else
    if(r!=1.0){ r=1.0/sqrt(r); printf(" normvec 1-r*(1/r)=%.1e",1-r*(1/r));
      for(i=0;i<B_NUM;i++)for(j=0l;j<b_len[i];j++) /* call mltly_av() ??? */
        v->v[i][j]=M_d2f(M_f2d(v->v[i][j])*r);
      for(r=i=0;i<B_NUM;i++)for(j=0l;j<b_len[i];j++)r+=Norm(M_f2d(v->v[i][j]));
      printf(" 1-norm=%.1e",1-r);
    }
    return r;
}

/* if blk==-1 save all blocks, idx>=0 index for name="d%d.vec" */
int savevec(tvec *v,const char *nam,int idx,int blk){
    FILE *f;ULong n2;int i=0;char s1[128],*s2;
    if(idx>=0) i=snprintf(s2=s1,128,nam,idx); else s2=(char *)nam;
    if(i<0) E0("file name to long");
    if(vvv&1)printf(" save_v %s",s2);
    if(!(f=fopen(s2,"wb")) ){ E1("fopen(%s)",s2);return 1;}
    for(i=((blk<0)?0:blk);i<((blk<0)?B_NUM:blk+1);i++){
      if(b_len[i]==0) continue;
      if(!v->v[i]){ E1(" failed, v[%d]==0",i); return 3;}
      if(blk>=0)if(fseek(f,(long)b_ofs[blk],SEEK_SET))E0("fseek");
      n2=fwrite((void *)v->v[i],Csize,(size_t)b_len[i],f);
      if( n2!=b_len[i] ){ E0("fwrite");return 2; }
    } fclose(f);return 0;}
int loadvec(tvec *v,const char *nam,int idx,int blk){
    FILE *f;ULong n2;int i=0;char s1[128],*s2;
    if(idx>=0) i=snprintf(s2=s1,128,nam,idx); else s2=(char *)nam;
    if(i<0) E0("file name to long");
    if(vvv&1)printf(" load_v %s",s2);
    if(!(f=fopen(s2,"rb")) ){ E0("fopen");return 1;}
    for(i=((blk<0)?0:blk);i<((blk<0)?B_NUM:blk+1);i++){
      if(b_len[i]==0) continue;
      if(!v->v[i]){ E1(" failed, v[%d]==0",i); return 3;}
      if(blk>=0)if(fseek(f,(long)b_ofs[blk],SEEK_SET))E0("fseek");
      n2=fread((void *)v->v[i],Csize,(size_t)b_len[i],f);
      if( n2!=b_len[i] ){ E0("fread");return 2; }
    } fclose(f);return 0;}
int loadvec2(tvec *v,const char *nam,int idx){         /* lade vec von textfile */
    FILE *f;double r;ULong j;tbase i;int k,l,blk,ret;char buf[255];
    int ii=0; char s1[128],*s2;
    if(idx>=0) ii=snprintf(s2=s1,128,nam,idx); else s2=(char *)nam;
    if(ii<0) E0("file name to long");
    if(!v){ printf(" failed, v==NULL"); return 3;}
    if(!(f=fopen(s2,"rt"))){E1("loadvec_fopen(%s)",s2);return 1;}
    clrvec(v);ret=1;
    while(!feof(f)){
        if(!fgets(buf,255,f))break; /* zeile lesen: 0.25 |ud.3u> /n */
        if(sscanf(buf,"%lf",&r)!=1)continue;
        for(k=0;buf[k];k++) if(buf[k]=='|')break;
        i=0l;j=0l;l=2*(nn-1) CM(/2,+0,+0);
        if(buf[k]=='|'){
           for(k++;buf[k]!=0 && l>=0;k++,CM(l--,l-=2,l-=2))	/* Apr99 */
           if(buf[k]=='.'){                   } else
           if(buf[k]=='u'){i|=CM(0l,1l,1l)<<l;} else
           if(buf[k]=='d'){i|=CM(1l,2l,2l)<<l;} else
           if(buf[k]=='3'){i|=         3l <<l;} else
           E0("unknown symbol (not .ud3)"); }
        else E1("no config |xxxx> found in %s",nam);   
        j=b2i(i); blk=b_getnum(j);   
        if(vvv)printf("\n read: %6.3f |%s> => %ld",r,ckfg(i),j);
        v->v[blk][j-b_ofs[blk]]=M_d2f(r);ret=0;}
    fclose(f);return ret;}
   
#ifndef RAND_MAX
#define RAND_MAX 0x7fff
#endif
extern int startvec;
/* --- last change: 7.12.98 JS bug! --- 
 *  sum v[] during setting for faster normalization???
 *  ini could be choosen randomly to ensure different numeric errors
 */
void randvec(tvec *v,int ini){
    static thbuf *hbuf=NULL; double ee,emin=9e6; int en=0;
    ULong i; int blk;
    ini+=(startvec&~7);
    if(vvv) printf(" randvec%d-%d",startvec&7,ini);
    if(startvec==1){ clrvec(v);blk=b_getnum(ini-1l);
      v->v[blk][(ULong)ini-1l-b_ofs[blk]]=M_d2f(1); } else
    if(startvec==2){ loadvec(v,FileGS,ini,-1); } else
    if(startvec==3){ loadvec2(v,FileSV,ini); } else
    if(startvec==5){
      printf(" lowestHdiag");  /* best for algo==2 ??? was Ising until v2.25 */
      if(!hbuf)hbuf=alloc_hbuf();
      for(blk=0;blk<B_NUM;blk++) for(i=0l;i<b_len[blk];i++){
        hamilton_nhv(b_ofs[blk]+i,NULL,hbuf,0);
        ee=hbuf->nhn;
        /* <n|Hn> Diagonal-H von |n> */
        if(ee< emin) { emin=ee; en=0; }
        if(ee==emin) en++;
      }
      printf(" emin=%g EA=%d",emin,en);
      for(blk=0;blk<B_NUM;blk++) for(i=0l;i<b_len[blk];i++){
        hamilton_nhv(b_ofs[blk]+i,NULL,hbuf,0);
        ee=hbuf->nhn;
         /* <n|Hn> Diagonal-H von |n> */
        v->v[blk][i]=M_d2f( ((ee==emin)?1:0) );
        if (ee==emin) {blk=B_NUM;break;}
      }
    } else
    if(startvec>4){ printf(" no startvec"); } else
    { /* srand(rnd);rnd++; -- bad idea, resulting in a bug until v2.17 */ 
      /* ToDo: spread in a way that norm is exactly 1.0, use l^2+m^2=n^2 */
      srand(ini);
      for(blk=0;blk<B_NUM;blk++) for(i=0l;i<b_len[blk];i++)
        v->v[blk][i]=M_d2f((1023 & rand())/1024.); /* Jul00 */
      v->v[0][0]=M_d2f(0.5);	/* be sure v is never zero */
    }
    normvec(v); return;
}
    
cplx sprod(tvec *a, tvec *b){ /* scalar product REIHENFOLGE!!!! */
    cplx r=0.0;ULong i;int blk;
    for(blk=0;blk<B_NUM;blk++)
    for(i=0l;i<b_len[blk];i++)r+=Conj(M_f2d(a->v[blk][i])) * M_f2d(b->v[blk][i]);
    return(r);
}
    
int  iortho(tvec *v0, tvec *v1){ /* 0=ok,1=parallel */
    cplx r1;zahl r2;ULong i;int blk;
    r1=sprod(v0,v1);
 /* if(fabs(r1)>0.99999){printf(" iortho: v0=v1 ");return(1);} */
    if(Norm(r1)<1.0e-10){return(0);}
    r2=0.0;
    for(blk=0;blk<B_NUM;blk++)for(i=0l;i<b_len[blk];i++){
      /* Conj(r1) for correct degeneracy for complex vectors */
      v0->v[blk][i]-=M_d2f(Conj(r1)*M_f2d(v1->v[blk][i])); /* round error ??? */
      r2+=Norm(M_f2d(v0->v[blk][i]));
    }
    /* 0.995 was successfull for degeneracy of kagome-JJ N=54 u+d=51+3 q_0
     * 0.99999 results in to high degeneracy (909 instead of 72) Jul04
     * */
    if(Norm(r1)>0.995){printf(" iortho: v0=v1");return(1);}
    r2=1.0/sqrt(r2);
    for(blk=0;blk<B_NUM;blk++)for(i=0l;i<b_len[blk];i++)
      v0->v[blk][i]=M_d2f(M_f2d(v0->v[blk][i])*r2);  /* call mltly_av ??? */
    return(0); }

